import axinite as ax
import axinite.tools as axtools
import json
from numba import jit

def load(args: axtools.AxiniteArgs, path: str = "", dont_change_args: bool = False, verbose: bool = False):
    """Preloads a simulation.

    Args:
        args (AxiniteArgs): An AxiniteArgs object containing simulation parameters.
        path (str, optional): The path to dump the computed simulation to. Defaults to "", which skips data dumping.
        dont_change_args (bool, optional): True if the load function should take care not to edit the args object. Defaults to False.
        verbose (bool, optional): Whether the load function should print to console. Defaults to False.

    Returns:
        list[axtools.Body]: A list of Body objects containing the preloaded simulation data.
    """
    
    if args.action == None: 
        ntotal = args.limit / args.delta

        @jit(nopython=False)
        def default_action(bodies, t, limit, delta, n): print("Timestep", n, "/", ntotal, "(", round(t / limit * 100, 2), "\b% )\033[F")

        args.action = default_action if verbose else None
        args.action_frequency = 200
    if args.backend == None: args.backend = ax.verlet_backend
    if verbose: print(f"Initializing system...")
    bodies = ax.load(*args.unpack(), t=args.t, modifier=args.modifier, action=args.action, action_frequency=args.action_frequency)
    if verbose: 
        print(f"Finished with {bodies[0]._inner['r'].shape[0]} timesteps \033[K")
        print(f"Synchronizing inner and outer bodies...")

    _bodies = []
    for i, body in enumerate(bodies):
        _bodies.append(axtools.Body(body.name, body.mass, args.limit, args.delta, position=body.r(0), velocity=body.v(0)))
        _bodies[i].retain = args.bodies[i].retain
        _bodies[i].color = args.bodies[i].color
        _bodies[i].light = args.bodies[i].light
        _bodies[i].radius = args.bodies[i].radius
        _bodies[i].radius_multiplier = args.bodies[i].radius_multiplier

        _bodies[i]._inner = body._inner
    
    if not dont_change_args:
        args.t = args.limit
        args.bodies = _bodies
    if path == "":
        return _bodies
    else: 
        if verbose: print(f"Saving system data to file...")
        axtools.save(args, path)
        if verbose: print(f"Finished saving system data to file")
        return _bodies


import numpy as np
import axinite as ax

class Body:
    """
    A class that represents a body in the simulation.

    Attributes:
        name (str): The body's name.
        mass (np.float64): The mass of the body in kilograms.
        position (np.ndarray): The initial position of the body (in vector form).
        velocity (np.ndarray): The initial velocity of the body (in vector form).
    """

    def __init__(self, name: str, mass: np.float64, limit: np.float64, delta: np.float64, position: np.ndarray = None, velocity: np.ndarray = None):
        """
        Initializes a new Body object.

        Args:
            name (str): The body's name.
            mass (np.float64): The mass of the body in kilograms.
            limit (np.float64): The length of the simulation in seconds.
            delta (np.float64): The frequency at which the simulation should be computed in seconds.
            position (np.ndarray, optional): The initial position of the body (in vector form). Defaults to None.
            velocity (np.ndarray, optional): The initial velocity of the body (in vector form). Defaults to None.
        """
        if position is not None:
            position = np.array(position, dtype=np.float64)
        if velocity is not None:
            velocity = np.array(velocity, dtype=np.float64)
        mass = np.float64(mass)
        limit = np.float64(limit)
        delta = np.float64(delta)
        
        self.mass = mass
        "The mass of the object in kilograms."

        self.name = name
        "The name of the object."

        self._inner = ax._body(limit, delta, name, mass)

        if position is not None: self._inner["r"][0] = position
        if velocity is not None: self._inner["v"][0] = velocity

        self._inner["n"] = name
        self._inner["m"] = mass

    def r(self, t: np.float64) -> np.ndarray:
        """Returns the position of the object at a specific time.

        Args:
            t (np.float64): The time to get the position at.

        Returns:
            np.ndarray: The position of the object at the time.
        """
        return self._inner["r"][int(t)]

    def v(self, t: np.float64) -> np.ndarray:
        """Returns the velocity of the object at a specific time.

        Args:
            t (np.float64): The time to get the velocity at.

        Returns:
            np.ndarray: The velocity of the object at the time.
        """
        return self._inner["v"][int(t)]

    @property
    def rs(self) -> np.ndarray: return self._inner["r"]
    @property
    def vs(self) -> np.ndarray: return self._inner["v"]


import axinite as ax
import axinite.analysis as axana
from numba import jit
import numpy as np

class Orbit:
    def __init__(self, central: ax.Body, satellite: ax.Body):
        self.central = central
        "The central body of the orbit."

        self.satellite = satellite
        "The satellite of the orbit."

        self.apogee, self.perigee = self._apogee_perigee()
        "The apogee/perigee of the orbit."

        self.eccentricity = self._eccentricity()
        "The eccentricity of the orbit."

        self.inclination = self._inclination()
        "The inclination of the orbit in radians."

        self.inclination_deg = np.degrees(self.inclination)
        "The inclination of the orbit in degrees."

        self.semi_major_axis = self._semi_major_axis()
        "The semi-major axis of the orbit."

        self.orbital_period = self._orbital_period()
        "The orbital period of the orbit in seconds."

        self.orbital_velocity = self._orbital_velocity()
        "The orbital velocity of the orbit in meters per second."

    def _apogee_perigee(self):
        relative = self.satellite._inner["r"] - self.central._inner["r"]
        absolute = np.linalg.norm(relative, axis=1)
        max_index = np.argmax(absolute)
        min_index = np.argmin(absolute)
        apogee = absolute[max_index]
        perigee = absolute[min_index]
        return (apogee, perigee)
    
    def _eccentricity(self):
        apogee_magnitude = np.linalg.norm(self.apogee)
        perigee_magnitude = np.linalg.norm(self.perigee)
        return (apogee_magnitude - perigee_magnitude) / (apogee_magnitude + perigee_magnitude)
    
    def _inclination(self):
        relative = self.satellite._inner["r"] - self.central._inner["r"]
        z = relative[:, 2]
        r = np.linalg.norm(relative, axis=1)
        return np.arccos(z / r)

    def _semi_major_axis(self):
        apogee_magnitude = np.linalg.norm(self.apogee)
        perigee_magnitude = np.linalg.norm(self.perigee)
        return (apogee_magnitude + perigee_magnitude) / 2
    
    def _orbital_period(self):
        total_mass = self.central.mass + self.satellite.mass
        return 2 * np.pi * np.sqrt(self.semi_major_axis**3 / (total_mass * ax.G))

    def _orbital_velocity(self):
        return 2 * np.pi * self.semi_major_axis / self._orbital_period()
    


import axinite as ax

def modifier_from(functions: list[tuple['function', 'function']]) -> 'function':
    def modifier(body, f, bodies, t, delta, limit, n):
        for function in functions:
            if function[0](body, f, bodies, t, delta, limit, n):
                f = function[1](body, f, bodies, t, delta, limit, n)
        return f
    return modifier


import numpy as np
from numba import jit
import axinite as ax

def verlet_nojit_backend(delta, limit, bodies, action=None, modifier=None, t=0.0, action_frequency=200):
    """A integration backend for the Verlet method without JIT.

    Args:
        delta (np.float64): The change in time between each step.
        limit (np.float64): The maximum time of the simulation.
        bodies (np.ndarray): An array of bodies.
        action (function, optional): A function to be called with frequency action_frequency. Defaults to None.
        modifier (function, optional): A function to be called at every timestep to modify the forces on the bodies. Defaults to None.
        t (float, optional): The initial starting time of the simulation. Defaults to -1.0.
        action_frequency (int, optional): The frequency at which to call the action. Defaults to 200.

    Returns:
        np.ndarray: The bodies after the simulation.
    """

    if t is None:
        t = 0.0
    if t > 0.0:
        raise Exception("Verlet method does not support non-zero initial time.")
    t = 0.0 + delta 
    n = 1

    for i, body in enumerate(bodies):
        f = np.zeros(3)
        for j, other in enumerate(bodies):
            if i != j: f += ax.gravitational_force_jit(body["m"], other["m"], body["r"][0] - other["r"][0])
        if modifier is not None: f = modifier(body, f, bodies=bodies, t=t, delta=delta, limit=limit, n=n)
        body["r"][1] = body["r"][0] + body["v"][0] * delta + 0.5 * (f / body["m"]) * delta**2
        body["v"][1] = (body["r"][1] - body["r"][0]) / (2 * delta)
    
    n += 1
    t += delta

    while t < limit:
        for i, body in enumerate(bodies):
            f = np.zeros(3)
            for j, other in enumerate(bodies):
                if i != j: f += ax.gravitational_force_jit(body["m"], other["m"], body["r"][n-1] - other["r"][n-1])
            if modifier is not None: f = modifier(body, f, bodies=bodies, t=t, delta=delta, limit=limit, n=n)

            body["r"][n] = body["r"][n-1] * 2 - body["r"][n-2] + (f / body["m"]) * delta**2
            body["v"][n] = (body["r"][n] - body["r"][n-1]) / (2 * delta)
        if action is not None and n % action_frequency == 0: action(bodies, t, limit=limit, delta=delta, n=n)
        n += 1
        t += delta
    
    return bodies

def verlet_backend(delta, limit, bodies, action=None, modifier=None, t=0.0, action_frequency=200):
    """A integration backend for the Verlet method with JIT.

    Args:
        delta (np.float64): The change in time between each step.
        limit (np.float64): The maximum time of the simulation.
        bodies (np.ndarray): An array of bodies.
        action (function, optional): A function to be called with frequency action_frequency. Defaults to None.
        modifier (function, optional): A function to be called at every timestep to modify the forces on the bodies. Defaults to None.
        t (float, optional): The initial starting time of the simulation. Defaults to 0.0.
        action_frequency (int, optional): The frequency at which to call the action. Defaults to 200.

    Returns:
        np.ndarray: The bodies after the simulation.
    """

    compiled = jit(verlet_nojit_backend, nopython=False)
    return compiled(delta, limit, bodies, action=action, modifier=modifier, t=t, action_frequency=action_frequency)